#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Dict, Optional

import bioio_base as biob
import dask.array as da
import numpy as np
from fsspec.implementations.local import LocalFileSystem
from imageio import get_writer

from .two_d_writer import TwoDWriter
from .writer import Writer

###############################################################################


class TimeseriesWriter(Writer):
    """
    A writer for timeseries Greyscale, RGB, or RGBA image data.
    Primarily directed at formats: "gif", "mp4", "mkv", etc.
    """

    _TIMEPOINT_DIMENSIONS = [
        biob.dimensions.DimensionNames.Time,
        biob.dimensions.DimensionNames.SpatialY,
        biob.dimensions.DimensionNames.SpatialX,
    ]
    _TIMEPOINT_WITH_SAMPLES_DIMENSIONS = _TIMEPOINT_DIMENSIONS + [
        biob.dimensions.DimensionNames.Samples
    ]

    DIM_ORDERS = {
        3: "".join(_TIMEPOINT_DIMENSIONS),  # Greyscale
        4: "".join(_TIMEPOINT_WITH_SAMPLES_DIMENSIONS),  # RGB / RGBA
    }

    @staticmethod
    def _write_chunks(
        f: str,
        extension: str,
        imageio_mode: str,
        fps: int,
        data: da.Array,
        dim_order: str,
    ) -> None:
        with get_writer(
            f,
            format=extension,
            mode=imageio_mode,
            fps=fps,
        ) as writer:
            # Make each chunk of the dask array be a frame
            chunks = tuple(
                1 if dim == biob.dimensions.DimensionNames.Time else -1
                for dim in dim_order
            )
            data = data.rechunk(chunks)

            # Save each frame
            for block in data.blocks:
                # Need to squeeze to remove the singleton T dimension
                writer.append_data(da.squeeze(block).compute())

    @staticmethod
    def save(
        data: biob.types.ArrayLike,
        uri: biob.types.PathLike,
        dim_order: Optional[str] = None,
        fps: int = 24,
        fs_kwargs: Dict[str, Any] = {},
        **kwargs: Any,
    ) -> None:
        """
        Write a data array to a file.

        Parameters
        ----------
        data: biob.types.ArrayLike
            The array of data to store. Data must have either three or four dimensions.
        uri: biob.types.PathLike
            The URI or local path for where to save the data.
        dim_order: str
            The dimension order of the provided data.
            Default: None. Based off the number of dimensions, will assume
            the dimensions -- three dimensions: TYX and four dimensions: TYXS.
        fps: int
            Frames per second to attach as metadata.
            Default: 24
        fs_kwargs: Dict[str, Any]
            Any specific keyword arguments to pass down to the fsspec created
            filesystem.
            Default: {}

        Examples
        --------
        Data is the correct shape and dimension order

        >>> image = dask.array.random.random((50, 100, 100))
        ... TimeseriesWriter.save(image, "file.gif")

        Data provided with current dimension order

        >>> image = numpy.random.rand(100, 3, 1024, 2048)
        ... TimeseriesWriter.save(image, "file.mkv", "TSYX")

        Save to remote

        >>> image = numpy.random.rand(300, 100, 100, 3)
        ... TimeseriesWriter.save(image, "s3://my-bucket/file.png")

        Raises
        ------
        IOError
            Cannot write FFMPEG formats to remote storage.

        Notes
        -----
        This writer can also be useful when wanting to create a timeseries image using
        a non-time dimension. For example, creating a timeseries image where each frame
        is a Z-plane from a source volumetric image as seen below.

        >>> image = BioImage("some_z_stack.ome.tiff")
        ... TimeseriesWriter.save(
        ...     data=image.get_image_data("ZYX", T=0, C=0),
        ...     uri="some_z_stack.mp4",
        ...     # Overloading the Z dimension as the Time dimension
        ...     # Technically not needed as it would have been assumed due to three dim
        ...     dim_order="TYX",
        ... )

        """
        # Check unpack uri and extension
        fs, path = biob.io.pathlike_to_fs(uri, fs_kwargs=fs_kwargs)
        (
            extension,
            imageio_mode,
        ) = TwoDWriter.get_extension_and_mode(path)

        # Convert to dask array to make downstream usage of data have a consistent API
        if isinstance(data, np.ndarray):
            data = da.from_array(data)

        # Shorthand num dimensions
        n_dims = len(data.shape)

        # Check num dimensions
        if n_dims not in TimeseriesWriter.DIM_ORDERS:
            raise biob.exceptions.UnexpectedShapeError(
                f"TimeseriesWriter requires that data must have either 3 or 4 "
                f"dimensions. Provided data with {n_dims} dimensions. ({data.shape})"
            )

        # Assume dim order if not provided
        if dim_order is None:
            dim_order = TimeseriesWriter.DIM_ORDERS[n_dims]

        # Uppercase dim order
        dim_order = dim_order.upper()

        # Check dimensions provided in the dim order string are all T, Y, X, or S
        if any(
            [
                dim not in TimeseriesWriter._TIMEPOINT_WITH_SAMPLES_DIMENSIONS
                for dim in dim_order
            ]
        ):
            raise biob.exceptions.InvalidDimensionOrderingError(
                f"The dim_order parameter only accepts dimensions: "
                f"{TimeseriesWriter._TIMEPOINT_WITH_SAMPLES_DIMENSIONS}. "
                f"Provided dim_order string: '{dim_order}'."
            )

        # Transpose dimensions if dim_order not ready for imageio
        if dim_order != TimeseriesWriter.DIM_ORDERS[n_dims]:
            # Actual reshape of the data
            data = biob.transforms.reshape_data(
                data,
                given_dims=dim_order,
                return_dims=TimeseriesWriter.DIM_ORDERS[n_dims],
            )

            # Set dim order to updated order
            dim_order = TimeseriesWriter.DIM_ORDERS[n_dims]

        # Handle FFMPEG formats
        if extension in TwoDWriter.FFMPEG_FORMATS:
            # FFMPEG can only handle local files
            # https://github.com/imageio/imageio-ffmpeg/issues/28#issuecomment-566012783
            if not isinstance(fs, LocalFileSystem):
                raise IOError(
                    f"Can only write to local files for formats: "
                    f"{TwoDWriter.FFMPEG_FORMATS}."
                )

            # Else, write with local
            TimeseriesWriter._write_chunks(
                f=path,
                extension=extension,
                imageio_mode=imageio_mode,
                fps=fps,
                data=data,
                dim_order=dim_order,
            )

        # Handle all non-ffmpeg formats
        else:
            with fs.open(path, "wb") as open_resource:
                TimeseriesWriter._write_chunks(
                    f=open_resource,
                    extension=extension,
                    imageio_mode=imageio_mode,
                    fps=fps,
                    data=data,
                    dim_order=dim_order,
                )



import math
from typing import Dict, List, Optional, Tuple

import bioio_base as biob
import zarr
from ome_zarr.io import parse_url
from ome_zarr.scale import Scaler
from ome_zarr.writer import write_image
from zarr.storage import default_compressor

from ..ome_utils import generate_ome_channel_id


class OmeZarrWriter:
    def __init__(self, uri: biob.types.PathLike):
        """
        Constructor.

        Parameters
        ----------
        uri: biob.types.PathLike
            The URI or local path for where to save the data.
        """
        # Resolve final destination
        fs, path = biob.io.pathlike_to_fs(uri)

        # Save image to zarr store!
        self.store = parse_url(uri, mode="w").store
        self.root_group = zarr.group(store=self.store)

    @staticmethod
    def build_ome(
        size_z: int,
        image_name: str,
        channel_names: List[str],
        channel_colors: List[int],
        channel_minmax: List[Tuple[float, float]],
    ) -> Dict:
        """
        Create the omero metadata for an OME zarr image

        Parameters
        ----------
        size_z:
            Number of z planes
        image_name:
            The name of the image
        channel_names:
            The names for each channel
        channel_colors:
            List of all channel colors
        channel_minmax:
            List of all (min, max) pairs of channel intensities

        Returns
        -------
        Dict
            An "omero" metadata object suitable for writing to ome-zarr
        """
        ch = []
        for i in range(len(channel_names)):
            ch.append(
                {
                    "active": True,
                    "coefficient": 1,
                    "color": f"{channel_colors[i]:06x}",
                    "family": "linear",
                    "inverted": False,
                    "label": channel_names[i],
                    "window": {
                        "end": float(channel_minmax[i][1]),
                        "max": float(channel_minmax[i][1]),
                        "min": float(channel_minmax[i][0]),
                        "start": float(channel_minmax[i][0]),
                    },
                }
            )

        omero = {
            "id": 1,  # ID in OMERO
            "name": image_name,  # Name as shown in the UI
            "version": "0.4",  # Current version
            "channels": ch,
            "rdefs": {
                "defaultT": 0,  # First timepoint to show the user
                "defaultZ": size_z // 2,  # First Z section to show the user
                "model": "color",  # "color" or "greyscale"
            },
            # TODO: can we add more metadata here?
            # # from here down this is all extra and not part of the ome-zarr spec
            # "meta": {
            #     "projectDescription": "20+ lines of gene edited cells etc",
            #     "datasetName": "aics_hipsc_v2020.1",
            #     "projectId": 2,
            #     "imageDescription": "foo bar",
            #     "imageTimestamp": 1277977808.0,
            #     "imageId": 12,
            #     "imageAuthor": "danielt",
            #     "imageName": "AICS-12_143.ome.tif",
            #     "datasetDescription": "variance dataset after QC",
            #     "projectName": "aics cell variance project",
            #     "datasetId": 3
            # },
        }
        return omero

    @staticmethod
    def _build_chunk_dims(
        chunk_dim_map: Dict[str, int],
        dimension_order: str = biob.dimensions.DEFAULT_DIMENSION_ORDER,
    ) -> Tuple[int, ...]:
        return tuple(chunk_dim_map[d] for d in dimension_order)

    def write_image(
        self,
        # TODO how to pass in precomputed multiscales?
        image_data: biob.types.ArrayLike,  # must be 3D, 4D or 5D
        image_name: str,
        physical_pixel_sizes: Optional[biob.types.PhysicalPixelSizes],
        channel_names: Optional[List[str]],
        channel_colors: Optional[List[int]],
        chunk_dims: Optional[Tuple] = None,
        scale_num_levels: int = 1,
        scale_factor: float = 2.0,
        dimension_order: Optional[str] = None,
    ) -> None:
        """
        Write a data array to a file.
        NOTE that this API is not yet finalized and will change in the future.

        Parameters
        ----------
        image_data: biob.types.ArrayLike
            The array of data to store. Data arrays must have 2 to 6 dimensions. If a
            list is provided, then it is understood to be multiple images written to the
            ome-tiff file. All following metadata parameters will be expanded to the
            length of this list.
        image_name: str
            string representing the name of the image
        physical_pixel_sizes: Optional[biob.types.PhysicalPixelSizes]
            PhysicalPixelSizes object representing the physical pixel sizes in Z, Y, X
            in microns.
            Default: None
        channel_names: Optional[List[str]]
            Lists of strings representing the names of the data channels
            Default: None
            If None is given, the list will be generated as a 0-indexed list of strings
            of the form "Channel:image_index:channel_index"
        channel_colors: Optional[List[int]]
            List of rgb color values per channel or a list of lists for each image.
            These must be values compatible with the OME spec.
            Default: None
        scale_num_levels: Optional[int]
            Number of pyramid levels to use for the image.
            Default: 1 (represents no downsampled levels)
        scale_factor: Optional[float]
            The scale factor to use for the image. Only active if scale_num_levels > 1.
            Default: 2.0
        dimension_order: Optional[str]
            The dimension order of the data. If None is given, the dimension order will
            be guessed from the number of dimensions in the data according to TCZYX
            order.

        Examples
        --------
        Write a TCZYX data set to OME-Zarr

        >>> image = numpy.ndarray([1, 10, 3, 1024, 2048])
        ... writer = OmeZarrWriter("/path/to/file.ome.zarr")
        ... writer.write_image(image)

        Write multi-scene data to OME-Zarr, specifying channel names

        >>> image0 = numpy.ndarray([3, 10, 1024, 2048])
        ... image1 = numpy.ndarray([3, 10, 512, 512])
        ... writer = OmeZarrWriter("/path/to/file.ome.zarr")
        ... writer.write_image(image0, "Image:0", ["C00","C01","C02"])
        ... writer.write_image(image1, "Image:1", ["C10","C11","C12"])
        """
        ndims = len(image_data.shape)
        if ndims < 2 or ndims > 5:
            raise biob.exceptions.InvalidDimensionOrderingError(
                f"Image data must have 2, 3, 4, or 5 dimensions. "
                f"Received image data with shape: {image_data.shape}"
            )
        if dimension_order is None:
            dimension_order = biob.dimensions.DEFAULT_DIMENSION_ORDER[-ndims:]
        if len(dimension_order) != ndims:
            raise biob.exceptions.InvalidDimensionOrderingError(
                f"Dimension order {dimension_order} does not match data "
                f"shape: {image_data.shape}"
            )
        if (
            len(set(dimension_order) - set(biob.dimensions.DEFAULT_DIMENSION_ORDER)) > 0
        ) or len(dimension_order) != len(set(dimension_order)):
            raise biob.exceptions.InvalidDimensionOrderingError(
                f"Dimension order {dimension_order} is invalid or contains"
                f"unexpected dimensions. Only {biob.dimensions.DEFAULT_DIMENSION_ORDER}"
                f"currently supported."
            )
        xdimindex = dimension_order.find(biob.dimensions.DimensionNames.SpatialX)
        ydimindex = dimension_order.find(biob.dimensions.DimensionNames.SpatialY)
        zdimindex = dimension_order.find(biob.dimensions.DimensionNames.SpatialZ)
        cdimindex = dimension_order.find(biob.dimensions.DimensionNames.Channel)
        if cdimindex > min(i for i in [xdimindex, ydimindex, zdimindex] if i > -1):
            raise biob.exceptions.InvalidDimensionOrderingError(
                f"Dimension order {dimension_order} is invalid. Channel dimension "
                f"must be before X, Y, and Z."
            )

        if chunk_dims is not None and len(chunk_dims) != ndims:
            raise biob.exceptions.UnexpectedShapeError(
                f"Chunk dimensions:{chunk_dims} do not match data. "
                f"Expected chunk dimension length:{ndims}"
            )

        if physical_pixel_sizes is None:
            pixelsizes = (1.0, 1.0, 1.0)
        else:
            pixelsizes = (
                physical_pixel_sizes.Z if physical_pixel_sizes.Z is not None else 1.0,
                physical_pixel_sizes.Y if physical_pixel_sizes.Y is not None else 1.0,
                physical_pixel_sizes.X if physical_pixel_sizes.X is not None else 1.0,
            )
        if channel_names is None:
            # TODO this isn't generating a very pretty looking name but it will be
            # unique
            channel_names = (
                [
                    generate_ome_channel_id(image_id=image_name, channel_id=i)
                    for i in range(image_data.shape[cdimindex])
                ]
                if cdimindex > -1
                else [generate_ome_channel_id(image_id=image_name, channel_id=0)]
            )
        if channel_colors is None:
            # TODO generate proper colors or confirm that the underlying lib can handle
            # None
            channel_colors = (
                [i for i in range(image_data.shape[cdimindex])]
                if cdimindex > -1
                else [0]
            )
        # Chunk spatial dimensions
        scale_dim_map = {
            biob.dimensions.DimensionNames.Time: 1.0,
            biob.dimensions.DimensionNames.Channel: 1.0,
            biob.dimensions.DimensionNames.SpatialZ: pixelsizes[0],
            biob.dimensions.DimensionNames.SpatialY: pixelsizes[1],
            biob.dimensions.DimensionNames.SpatialX: pixelsizes[2],
        }
        transforms = [
            [
                # the voxel size for the first scale level
                {
                    "type": "scale",
                    "scale": [scale_dim_map[d] for d in dimension_order],
                }
            ]
        ]
        # TODO precompute sizes for downsampled also.
        plane_size = (
            image_data.shape[xdimindex]
            * image_data.shape[ydimindex]
            * image_data.itemsize
        )

        target_chunk_size = 16 * (1024 * 1024)  # 16 MB
        # this is making an assumption of chunking whole XY planes.

        if chunk_dims is None:
            nplanes_per_chunk = int(math.ceil(target_chunk_size / plane_size))
            nplanes_per_chunk = (
                min(nplanes_per_chunk, image_data.shape[zdimindex])
                if zdimindex > -1
                else 1
            )
            chunk_dim_map = {
                biob.dimensions.DimensionNames.Time: 1,
                biob.dimensions.DimensionNames.Channel: 1,
                biob.dimensions.DimensionNames.SpatialZ: nplanes_per_chunk,
                biob.dimensions.DimensionNames.SpatialY: image_data.shape[ydimindex],
                biob.dimensions.DimensionNames.SpatialX: image_data.shape[xdimindex],
            }
            chunks = [
                dict(
                    chunks=OmeZarrWriter._build_chunk_dims(
                        chunk_dim_map=chunk_dim_map, dimension_order=dimension_order
                    ),
                    compressor=default_compressor,
                )
            ]
        else:
            chunks = [
                dict(
                    chunks=chunk_dims,
                    compressor=default_compressor,
                )
            ]

        lasty = image_data.shape[ydimindex]
        lastx = image_data.shape[xdimindex]
        # TODO scaler might want to use different method for segmentations than raw
        # TODO allow custom scaler or pre-computed multiresolution levels
        if scale_num_levels > 1:
            # TODO As of this writing, this Scaler is not the most general
            # implementation (it does things by xy plane) but it's code already
            # written that also works with dask, so it's a good starting point.
            scaler = Scaler()
            scaler.method = "nearest"
            scaler.max_layer = scale_num_levels - 1
            scaler.downscale = scale_factor if scale_factor is not None else 2
            for _ in range(scale_num_levels - 1):
                scale_dim_map[
                    biob.dimensions.DimensionNames.SpatialY
                ] *= scaler.downscale
                scale_dim_map[
                    biob.dimensions.DimensionNames.SpatialX
                ] *= scaler.downscale
                transforms.append(
                    [
                        {
                            "type": "scale",
                            "scale": [scale_dim_map[d] for d in dimension_order],
                        }
                    ]
                )

                if chunk_dims is None:
                    lasty = int(math.ceil(lasty / scaler.downscale))
                    lastx = int(math.ceil(lastx / scaler.downscale))
                    chunk_dim_map = {
                        biob.dimensions.DimensionNames.Time: 1,
                        biob.dimensions.DimensionNames.Channel: 1,
                    }
                    plane_size = lasty * lastx * image_data.itemsize
                    nplanes_per_chunk = int(math.ceil(target_chunk_size / plane_size))
                    nplanes_per_chunk = (
                        min(nplanes_per_chunk, image_data.shape[zdimindex])
                        if zdimindex > -1
                        else 1
                    )

                    chunk_dim_map[
                        biob.dimensions.DimensionNames.SpatialZ
                    ] = nplanes_per_chunk
                    chunk_dim_map[biob.dimensions.DimensionNames.SpatialY] = lasty
                    chunk_dim_map[biob.dimensions.DimensionNames.SpatialX] = lastx

                    chunks.append(
                        dict(
                            chunks=OmeZarrWriter._build_chunk_dims(
                                chunk_dim_map=chunk_dim_map,
                                dimension_order=dimension_order,
                            ),
                            compressor=default_compressor,
                        )
                    )
                else:
                    rescaley = int(math.ceil(chunk_dims[ydimindex] / scaler.downscale))
                    rescalex = int(math.ceil(chunk_dims[xdimindex] / scaler.downscale))
                    chunk_dims = tuple(list(chunk_dims[:-2]) + [rescaley, rescalex])

                    chunks.append(
                        dict(
                            chunks=chunk_dims,
                            compressor=default_compressor,
                        )
                    )

        else:
            scaler = None

        # try to construct per-image metadata
        ome_json = OmeZarrWriter.build_ome(
            image_data.shape[zdimindex] if zdimindex > -1 else 1,
            image_name,
            channel_names=channel_names,  # type: ignore
            channel_colors=channel_colors,  # type: ignore
            # This can be slow if computed here.
            # TODO: Rely on user to supply the per-channel min/max.
            channel_minmax=[
                (0.0, 1.0)
                for i in range(image_data.shape[cdimindex] if cdimindex > -1 else 1)
            ],
        )
        # TODO user supplies units?
        dim_to_axis = {
            biob.dimensions.DimensionNames.Time: {
                "name": "t",
                "type": "time",
                "unit": "millisecond",
            },
            biob.dimensions.DimensionNames.Channel: {"name": "c", "type": "channel"},
            biob.dimensions.DimensionNames.SpatialZ: {
                "name": "z",
                "type": "space",
                "unit": "micrometer",
            },
            biob.dimensions.DimensionNames.SpatialY: {
                "name": "y",
                "type": "space",
                "unit": "micrometer",
            },
            biob.dimensions.DimensionNames.SpatialX: {
                "name": "x",
                "type": "space",
                "unit": "micrometer",
            },
        }

        axes = [dim_to_axis[d] for d in dimension_order]

        # TODO image name must be unique within this root group
        group = self.root_group  # .create_group(image_name, overwrite=True)
        group.attrs["omero"] = ome_json

        write_image(
            image=image_data,
            group=group,
            scaler=scaler,
            axes=axes,
            # For each resolution, we have a List of transformation Dicts (not
            # validated). Each list of dicts are added to each datasets in order.
            coordinate_transformations=transforms,
            # Options to be passed on to the storage backend. A list would need to
            # match the number of datasets in a multiresolution pyramid. One can
            # provide different chunk size for each level of a pyramid using this
            # option.
            storage_options=chunks,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pathlib
from typing import Callable, Tuple

import bioio_base as biob
import imageio
import numpy as np
import pytest

from bioio.writers.timeseries_writer import TimeseriesWriter
from bioio.writers.two_d_writer import TwoDWriter

from ..conftest import array_constructor


@array_constructor
@pytest.mark.parametrize(
    "write_shape, write_dim_order, read_shape",
    [
        # TODO: Failing currently, needs work,
        # see https://github.com/bioio-devs/bioio/issues/10
        # ((30, 100, 100), None, (30, 100, 100)),
        # Note that files get saved out with RGBA, instead of just RGB
        ((30, 100, 100, 3), None, (30, 100, 100, 4)),
        # TODO: Failing currently, needs work,
        # see https://github.com/bioio-devs/bioio/issues/10
        # ((100, 30, 100), "XTY", (30, 100, 100)),
        # Note that files get saved out with RGBA, instead of just RGB
        ((3, 100, 30, 100), "SYTX", (30, 100, 100, 4)),
        pytest.param(
            (1, 1),
            None,
            None,
            marks=pytest.mark.xfail(raises=biob.exceptions.UnexpectedShapeError),
        ),
        pytest.param(
            (1, 1, 1, 1, 1),
            None,
            None,
            marks=pytest.mark.xfail(raises=biob.exceptions.UnexpectedShapeError),
        ),
        pytest.param(
            (1, 1, 1, 1, 1, 1),
            "STCZYX",
            None,
            marks=pytest.mark.xfail(raises=biob.exceptions.UnexpectedShapeError),
        ),
        pytest.param(
            (1, 1, 1, 1),
            "ABCD",
            None,
            marks=pytest.mark.xfail(
                raises=biob.exceptions.InvalidDimensionOrderingError
            ),
        ),
    ],
)
@pytest.mark.parametrize("filename", ["e.gif"])
def test_timeseries_writer(
    array_constructor: Callable,
    write_shape: Tuple[int, ...],
    write_dim_order: str,
    read_shape: Tuple[int, ...],
    filename: str,
    tmp_path: pathlib.Path,
) -> None:
    # Create array
    arr = array_constructor(write_shape, dtype=np.uint8)

    # Construct save end point
    save_uri = tmp_path / filename

    # Normal save
    TimeseriesWriter.save(arr, save_uri, write_dim_order)

    # TODO: Actually uncovered bug in DefaultReader :(
    # dask_a = aicsimageio.readers.default_reader.DefaultReader(save_uri).dask_data
    # data = dask_a.compute()
    # assert data.shape == read_shape
    # assert reader.shape[-1] <

    # Read written result and check basics
    fs, path = biob.io.pathlike_to_fs(save_uri)
    extension, mode = TwoDWriter.get_extension_and_mode(path)
    with fs.open(path) as open_resource:
        with imageio.get_reader(open_resource, format=extension, mode=mode) as reader:
            # Read and stack all frames
            frames = []
            for frame in reader:
                print(frame.shape)
                frames.append(frame)

            data = np.stack(frames)

            # Check basics
            assert data.shape == read_shape
            assert data.shape[-1] <= 4

            # Can't do "easy" testing because compression + shape mismatches on RGB data


@array_constructor
@pytest.mark.parametrize(
    "write_shape, write_dim_order, read_shape",
    [
        # We use 112 instead of 100 because FFMPEG block size warnings are annoying
        ((30, 112, 112), None, (30, 112, 112, 3)),
        # Note that files get saved out with RGBA, instead of just RGB
        ((30, 112, 112, 3), None, (30, 112, 112, 3)),
        ((112, 30, 112), "XTY", (30, 112, 112, 3)),
        # Note that files get saved out with RGBA, instead of just RGB
        ((3, 112, 30, 112), "SYTX", (30, 112, 112, 3)),
        pytest.param(
            (1, 1),
            None,
            None,
            marks=pytest.mark.xfail(raises=biob.exceptions.UnexpectedShapeError),
        ),
        pytest.param(
            (1, 1, 1, 1, 1),
            None,
            None,
            marks=pytest.mark.xfail(raises=biob.exceptions.UnexpectedShapeError),
        ),
        pytest.param(
            (1, 1, 1, 1, 1, 1),
            "STCZYX",
            None,
            marks=pytest.mark.xfail(raises=biob.exceptions.UnexpectedShapeError),
        ),
        pytest.param(
            (1, 1, 1, 1),
            "ABCD",
            None,
            marks=pytest.mark.xfail(
                raises=biob.exceptions.InvalidDimensionOrderingError
            ),
        ),
    ],
)
@pytest.mark.parametrize("filename", ["f.mp4"])
def test_timeseries_writer_ffmpeg(
    array_constructor: Callable,
    write_shape: Tuple[int, ...],
    write_dim_order: str,
    read_shape: Tuple[int, ...],
    filename: str,
    tmp_path: pathlib.Path,
) -> None:
    # Create array
    arr = array_constructor(write_shape, dtype=np.uint8)

    # Construct save end point
    save_uri = tmp_path / filename

    # Normal save
    TimeseriesWriter.save(arr, save_uri, write_dim_order)

    # Read written result and check basics
    fs, path = biob.io.pathlike_to_fs(save_uri)
    extension, mode = TwoDWriter.get_extension_and_mode(path)
    with fs.open(path) as open_resource:
        with imageio.get_reader(open_resource, format=extension, mode=mode) as reader:
            # Read and stack all frames
            frames = []
            for frame in reader:
                frames.append(frame)

            data = np.stack(frames)

            # Check basics
            assert data.shape == read_shape
            assert data.shape[-1] <= 4

            # Can't do "easy" testing because compression + shape mismatches on RGB data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pathlib
from typing import Callable, List, Optional, Tuple

import bioio_base as biob
import numpy as np
import pytest
from ome_zarr.io import parse_url
from ome_zarr.reader import Reader

from bioio.writers import OmeZarrWriter

from ..conftest import array_constructor


@array_constructor
@pytest.mark.parametrize(
    "write_shape, write_dim_order, expected_read_shape, expected_read_dim_order",
    [
        ((1, 2, 3, 4, 5), None, (1, 2, 3, 4, 5), "TCZYX"),
        ((1, 2, 3, 4, 5), "TCZYX", (1, 2, 3, 4, 5), "TCZYX"),
        ((2, 3, 4, 5, 6), None, (2, 3, 4, 5, 6), "TCZYX"),
        ((1, 1, 1, 1, 1), None, (1, 1, 1, 1, 1), "TCZYX"),
        ((5, 16, 16), None, (5, 16, 16), "ZYX"),
        ((5, 16, 16), "ZYX", (5, 16, 16), "ZYX"),
        ((5, 16, 16), "CYX", (5, 16, 16), "CYX"),
        ((5, 16, 16), "TYX", (5, 16, 16), "TYX"),
        pytest.param(
            (10, 5, 16, 16),
            "ZCYX",
            (10, 5, 16, 16),
            "ZCYX",
            marks=pytest.mark.xfail(
                raises=biob.exceptions.InvalidDimensionOrderingError
            ),
        ),
        ((5, 10, 16, 16), "CZYX", (5, 10, 16, 16), "CZYX"),
        ((15, 16), "YX", (15, 16), "YX"),
        pytest.param(
            (2, 3, 3),
            "AYX",
            None,
            None,
            marks=pytest.mark.xfail(
                raises=biob.exceptions.InvalidDimensionOrderingError
            ),
        ),
        ((2, 3, 3), "YXZ", (2, 3, 3), "YXZ"),
        pytest.param(
            (2, 5, 16, 16),
            "CYX",
            None,
            None,
            marks=pytest.mark.xfail(
                raises=biob.exceptions.InvalidDimensionOrderingError
            ),
        ),
        # error 6D data doesn't work yet
        pytest.param(
            (1, 2, 3, 4, 5, 3),
            None,
            None,
            None,
            marks=pytest.mark.xfail(
                raises=biob.exceptions.InvalidDimensionOrderingError
            ),
        ),
    ],
)
@pytest.mark.parametrize("filename", ["e.zarr"])
def test_ome_zarr_writer_dims(
    array_constructor: Callable,
    write_shape: Tuple[int, ...],
    write_dim_order: Optional[str],
    expected_read_shape: Tuple[int, ...],
    expected_read_dim_order: str,
    filename: str,
    tmp_path: pathlib.Path,
) -> None:
    # Create array
    arr = array_constructor(write_shape, dtype=np.uint8)

    # Construct save end point
    save_uri = tmp_path / filename

    # Normal save
    writer = OmeZarrWriter(save_uri)
    writer.write_image(arr, "", None, None, None, dimension_order=write_dim_order)

    # Read written result and check basics
    reader = Reader(parse_url(save_uri))
    node = list(reader())[0]
    num_levels = len(node.data)
    assert num_levels == 1
    level = 0
    shape = node.data[level].shape
    assert shape == expected_read_shape
    axes = node.metadata["axes"]
    dims = "".join([a["name"] for a in axes]).upper()
    assert dims == expected_read_dim_order


@array_constructor
@pytest.mark.parametrize(
    "write_shape, num_levels, scale, expected_read_shapes, expected_read_scales",
    [
        (
            (2, 4, 8, 16, 32),
            2,
            2,
            [(2, 4, 8, 16, 32), (2, 4, 8, 8, 16), (2, 4, 8, 4, 8)],
            [
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 2.0, 2.0],
                [1.0, 1.0, 1.0, 4.0, 4.0],
            ],
        ),
        (
            (16, 32),
            2,
            4,
            [(16, 32), (4, 8), (1, 2)],
            [
                [1.0, 1.0],
                [4.0, 4.0],
                [16.0, 16.0],
            ],
        ),
    ],
)
@pytest.mark.parametrize("filename", ["f.zarr"])
def test_ome_zarr_writer_scaling(
    array_constructor: Callable,
    write_shape: Tuple[int, ...],
    num_levels: int,
    scale: float,
    expected_read_shapes: List[Tuple[int, ...]],
    expected_read_scales: List[List[int]],
    filename: str,
    tmp_path: pathlib.Path,
) -> None:
    # Create array
    arr = array_constructor(write_shape, dtype=np.uint8)

    # Construct save end point
    save_uri = tmp_path / filename

    # Normal save
    writer = OmeZarrWriter(save_uri)
    writer.write_image(
        arr, "", None, None, None, scale_num_levels=num_levels, scale_factor=scale
    )

    # Read written result and check basics
    reader = Reader(parse_url(save_uri))
    node = list(reader())[0]
    read_num_levels = len(node.data)
    assert num_levels == read_num_levels
    print(node.metadata)
    for i in range(num_levels):
        shape = node.data[i].shape
        assert shape == expected_read_shapes[i]
        xforms = node.metadata["coordinateTransformations"][i]
        assert len(xforms) == 1
        assert xforms[0]["type"] == "scale"
        assert xforms[0]["scale"] == expected_read_scales[i]


@array_constructor
@pytest.mark.parametrize(
    "write_shape, chunk_dims, num_levels, expected_read_shapes",
    [
        (
            (2, 4, 8, 16, 32),
            (1, 1, 2, 16, 16),
            2,
            [(2, 4, 8, 16, 32), (2, 4, 8, 8, 16), (2, 4, 8, 4, 8)],
        ),
        (
            (16, 32),
            (2, 4),
            2,
            [(16, 32), (8, 16), (4, 8)],
        ),
    ],
)
@pytest.mark.parametrize("filename", ["e.zarr"])
def test_ome_zarr_writer_chunks(
    array_constructor: Callable,
    write_shape: Tuple[int, ...],
    chunk_dims: Tuple[int, ...],
    num_levels: int,
    filename: str,
    expected_read_shapes: List[Tuple[int, ...]],
    tmp_path: pathlib.Path,
) -> None:
    arr = array_constructor(write_shape, dtype=np.uint8)

    # Construct save end point

    baseline_save_uri = tmp_path / f"baseline_{filename}"
    save_uri = tmp_path / filename

    # Normal save
    writer = OmeZarrWriter(save_uri)
    writer.write_image(
        arr, "", None, None, None, chunk_dims=chunk_dims, scale_num_levels=num_levels
    )
    reader = Reader(parse_url(save_uri))
    node = list(reader())[0]

    # Check expected shapes
    for level in range(num_levels):
        shape = node.data[level].shape
        assert shape == expected_read_shapes[level]

    # Create baseline chunking to compare against manual.
    writer = OmeZarrWriter(baseline_save_uri)
    writer.write_image(arr, "", None, None, None, scale_num_levels=num_levels)
    reader_baseline = Reader(parse_url(baseline_save_uri))
    node_baseline = list(reader_baseline())[0]

    data = node.data[0]
    baseline_data = node_baseline.data[0]

    assert np.all(np.equal(data, baseline_data))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Dict, Optional, Tuple

import bioio_base as biob
import dask.array as da
from imageio import get_writer

from .writer import Writer

###############################################################################


class TwoDWriter(Writer):
    """
    A writer for image data is only 2 dimension with samples (RGB / RGBA) optional.
    Primarily directed at formats: "png", "jpg", etc.

    This is primarily a passthrough to imageio.imwrite.
    """

    _PLANE_DIMENSIONS = [
        biob.dimensions.DimensionNames.SpatialY,
        biob.dimensions.DimensionNames.SpatialX,
    ]
    _PLANE_WITH_SAMPLES_DIMENSIONS = _PLANE_DIMENSIONS + [
        biob.dimensions.DimensionNames.Samples
    ]

    DIM_ORDERS = {
        2: "".join(_PLANE_DIMENSIONS),  # Greyscale
        3: "".join(_PLANE_WITH_SAMPLES_DIMENSIONS),  # RGB / RGBA
    }
    FFMPEG_FORMATS = ["mov", "avi", "mpg", "mpeg", "mp4", "mkv", "wmv", "ogg"]

    @staticmethod
    def get_extension_and_mode(path: str) -> Tuple[str, str]:
        """
        Provided a path to a file, provided back the extension (format) of the file
        and the imageio read mode.

        Parameters
        ----------
        path: str
            The file to provide extension and mode info for.

        Returns
        -------
        extension: str
            The extension (a naive guess at the format) of the file.
        mode: str
            The imageio read mode to use for image reading.
        """
        # Select extension to handle special formats
        extension = path.split(".")[-1]

        # Set mode to many-image reading if FFMPEG format was provided
        # https://imageio.readthedocs.io/en/stable/userapi.html#imageio.get_reader
        if extension in TwoDWriter.FFMPEG_FORMATS:
            mode = "I"
        # Otherwise, have imageio infer the mode
        else:
            mode = "?"

        return extension, mode

    @staticmethod
    def save(
        data: biob.types.ArrayLike,
        uri: biob.types.PathLike,
        dim_order: Optional[str] = None,
        fs_kwargs: Dict[str, Any] = {},
        **kwargs: Any,
    ) -> None:
        """
        Write a data array to a file.

        Parameters
        ----------
        data: types.ArrayLike
            The array of data to store. Data must have either two or three dimensions.
        uri: types.PathLike
            The URI or local path for where to save the data.
        dim_order: str
            The dimension order of the provided data.
            Default: None. Based off the number of dimensions, will assume
            the dimensions similar to how
            https://github.com/bioio-devs/bioio-imageio/blob/main/bioio_imageio/reader.py
            reads in data. That is, two dimensions: YX and three dimensions: YXS.
        fs_kwargs: Dict[str, Any]
            Any specific keyword arguments to pass down to the fsspec created
            filesystem.
            Default: {}

        Examples
        --------
        Data is the correct shape and dimension order

        >>> image = dask.array.random.random((100, 100, 4))
        ... TwoDWriter.save(image, "file.png")

        Data provided with current dimension order

        >>> image = numpy.random.rand(3, 1024, 2048)
        ... TwoDWriter.save(image, "file.png", "SYX")

        Save to remote

        >>> image = numpy.random.rand(100, 100, 3)
        ... TwoDWriter.save(image, "s3://my-bucket/file.png")
        """
        # Check unpack uri and extension
        fs, path = biob.io.pathlike_to_fs(uri, fs_kwargs=fs_kwargs)
        (
            extension,
            imageio_mode,
        ) = TwoDWriter.get_extension_and_mode(path)

        # Assumption: if provided a dask array to save, it can fit into memory
        if isinstance(data, da.core.Array):
            data = data.compute()

        # Shorthand num dimensions
        n_dims = len(data.shape)

        # Check num dimensions
        if n_dims not in TwoDWriter.DIM_ORDERS:
            raise biob.exceptions.UnexpectedShapeError(
                f"TwoDWriter requires that data must have either 2 or 3 dimensions. "
                f"Provided data with {n_dims} dimensions. ({data.shape})"
            )

        # Assume dim order if not provided
        if dim_order is None:
            dim_order = TwoDWriter.DIM_ORDERS[n_dims]

        # Uppercase dim order
        dim_order = dim_order.upper()

        # Check dimensions provided in the dim order string are all Y, X, or S
        if any(
            [dim not in TwoDWriter._PLANE_WITH_SAMPLES_DIMENSIONS for dim in dim_order]
        ):
            raise biob.exceptions.InvalidDimensionOrderingError(
                f"The dim_order parameter only accepts dimensions: "
                f"{TwoDWriter._PLANE_WITH_SAMPLES_DIMENSIONS}. "
                f"Provided dim_order string: '{dim_order}'."
            )

        # Transpose dimensions if dim_order not ready for imageio
        if dim_order != TwoDWriter.DIM_ORDERS[n_dims]:
            data = biob.transforms.reshape_data(
                data, given_dims=dim_order, return_dims=TwoDWriter.DIM_ORDERS[n_dims]
            )

        # Save image
        with fs.open(path, "wb") as open_resource:
            with get_writer(
                open_resource,
                format=extension,
                mode=imageio_mode,
            ) as writer:
                writer.append_data(data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import abc
import typing

import bioio_base as biob

###############################################################################


class Writer(abc.ABC):
    """
    A small class to build standardized image writer functions.
    """

    @staticmethod
    @abc.abstractmethod
    def save(
        data: biob.types.ArrayLike,
        uri: biob.types.PathLike,
        dim_order: str = biob.dimensions.DEFAULT_DIMENSION_ORDER,
        **kwargs: typing.Any,
    ) -> None:
        """
        Write a data array to a file.

        Parameters
        ----------
        data: types.ArrayLike
            The array of data to store.
        uri: types.PathLike
            The URI or local path for where to save the data.
        dim_order: str
            The dimension order of the data.

        Examples
        --------
        >>> image = numpy.ndarray([1, 10, 3, 1024, 2048])
        ... DerivedWriter.save(image, "file.ome.tif", "TCZYX")

        >>> image = dask.array.ones((4, 100, 100))
        ... DerivedWriter.save(image, "file.png", "CYX")
        """
        # There are no requirements for n-dimensions of data.
        # The data provided can be 2D - ND.



import logging
from dataclasses import asdict, dataclass
from math import prod
from typing import Any, List, Tuple, Union

import dask.array as da
import numcodecs
import numpy as np
import skimage.transform
import zarr
from ngff_zarr.zarr_metadata import Axis, Dataset, Metadata, Scale, Translation
from zarr.storage import DirectoryStore, FSStore, default_compressor

from bioio import BioImage

log = logging.getLogger(__name__)

DimTuple = Tuple[int, int, int, int, int]

OME_NGFF_VERSION = "0.4"


def chunk_size_from_memory_target(
    shape: DimTuple, dtype: str, memory_target: int
) -> DimTuple:
    """
    Calculate chunk size from memory target in bytes.  The chunk size will be
    determined by considering a single T and C, and subdividing the remaining
    dims by 2 until the chunk fits within the size target.

    Parameters
    ----------
    shape:
        Shape of the array. Assumes a 5d TCZYX array.
    dtype:
        Data type of the array.
    memory_target:
        Memory target in bytes.

    Returns
    -------
    Chunk size tuple.
    """
    if len(shape) != 5:
        raise ValueError("shape must be a 5-tuple in TCZYX order")

    itemsize = np.dtype(dtype).itemsize
    # let's start by just mandating that chunks have to be no more than
    # 1 T and 1 C
    chunk_size = (1, 1, shape[2], shape[3], shape[4])
    while prod(chunk_size) * itemsize > memory_target:
        # chop every dim in half until they get down to 1
        chunk_size = (
            max(chunk_size[0] // 2, 1),
            max(chunk_size[1] // 2, 1),
            max(chunk_size[2] // 2, 1),
            max(chunk_size[3] // 2, 1),
            max(chunk_size[4] // 2, 1),
        )
    return chunk_size


def dim_tuple_to_dict(
    dims: Union[DimTuple, Tuple[float, float, float, float, float]]
) -> dict:
    if len(dims) != 5:
        raise ValueError("dims must be a 5-tuple in TCZYX order")
    return {"t": dims[0], "c": dims[1], "z": dims[2], "y": dims[3], "x": dims[4]}


def resize(
    image: da.Array, output_shape: Tuple[int, ...], *args: Any, **kwargs: Any
) -> da.Array:
    """
    Wrapped copy of "skimage.transform.resize"
    Resize image to match a certain size.

    Parameters
    ----------
    image: :class:`dask.array`
        The dask array to resize
    output_shape: tuple
        The shape of the resize array
    *args: list
        Arguments of skimage.transform.resize
    **kwargs: dict
        Keyword arguments of skimage.transform.resize

    Returns
    -------
    Resized image.
    """
    factors = np.array(output_shape) / np.array(image.shape).astype(float)
    # Rechunk the input blocks so that the factors achieve an output
    # blocks size of full numbers.
    better_chunksize = tuple(
        np.maximum(1, np.round(np.array(image.chunksize) * factors) / factors).astype(
            int
        )
    )
    image_prepared = image.rechunk(better_chunksize)

    # If E.g. we resize image from 6675 by 0.5 to 3337, factor is 0.49992509 so each
    # chunk of size e.g. 1000 will resize to 499. When assumbled into a new array, the
    # array will now be of size 3331 instead of 3337 because each of 6 chunks was
    # smaller by 1. When we compute() this, dask will read 6 chunks of 1000 and expect
    # last chunk to be 337 but instead it will only be 331.
    # So we use ceil() here (and in resize_block) to round 499.925 up to chunk of 500
    block_output_shape = tuple(
        np.ceil(np.array(better_chunksize) * factors).astype(int)
    )

    # Map overlap
    def resize_block(image_block: da.Array, block_info: dict) -> da.Array:
        # if the input block is smaller than a 'regular' chunk (e.g. edge of image)
        # we need to calculate target size for each chunk...
        chunk_output_shape = tuple(
            np.ceil(np.array(image_block.shape) * factors).astype(int)
        )
        return skimage.transform.resize(
            image_block, chunk_output_shape, *args, **kwargs
        ).astype(image_block.dtype)

    output_slices = tuple(slice(0, d) for d in output_shape)
    output = da.map_blocks(
        resize_block, image_prepared, dtype=image.dtype, chunks=block_output_shape
    )[output_slices]
    return output.rechunk(image.chunksize).astype(image.dtype)


def _pop_metadata_optionals(metadata_dict: dict) -> dict:
    for ax in metadata_dict["axes"]:
        if ax["unit"] is None:
            ax.pop("unit")

    if metadata_dict["coordinateTransformations"] is None:
        metadata_dict.pop("coordinateTransformations")

    return metadata_dict


def build_ome(
    size_z: int,
    image_name: str,
    channel_names: List[str],
    channel_colors: List[int],
    channel_minmax: List[Tuple[float, float]],
) -> dict:
    """
    Create the omero metadata for an OME zarr image

    Parameters
    ----------
    size_z:
        Number of z planes
    image_name:
        The name of the image
    channel_names:
        The names for each channel.  Must be of correct length!
    channel_colors:
        List of all channel colors
    channel_minmax:
        List of all (min, max) pairs of channel intensities

    Returns
    -------
    Dict
        An "omero" metadata object suitable for writing to ome-zarr
    """
    ch = []
    for i in range(len(channel_names)):
        ch.append(
            {
                "active": True,
                "coefficient": 1,
                "color": f"{channel_colors[i]:06x}",
                "family": "linear",
                "inverted": False,
                "label": channel_names[i],
                "window": {
                    "end": float(channel_minmax[i][1]),
                    "max": float(channel_minmax[i][1]),
                    "min": float(channel_minmax[i][0]),
                    "start": float(channel_minmax[i][0]),
                },
            }
        )

    omero = {
        "id": 1,  # ID in OMERO
        "name": image_name,  # Name as shown in the UI
        "version": OME_NGFF_VERSION,  # Current version
        "channels": ch,
        "rdefs": {
            "defaultT": 0,  # First timepoint to show the user
            "defaultZ": size_z // 2,  # First Z section to show the user
            "model": "color",  # "color" or "greyscale"
        },
        # TODO: can we add more metadata here?
        # Must check with the ngff spec.
        # # from here down this is all extra and not part of the ome-zarr spec
        # "meta": {
        #     "projectDescription": "20+ lines of gene edited cells etc",
        #     "datasetName": "aics_hipsc_v2020.1",
        #     "projectId": 2,
        #     "imageDescription": "foo bar",
        #     "imageTimestamp": 1277977808.0,
        #     "imageId": 12,
        #     "imageAuthor": "danielt",
        #     "imageName": "AICS-12_143.ome.tif",
        #     "datasetDescription": "variance dataset after QC",
        #     "projectName": "aics cell variance project",
        #     "datasetId": 3
        # },
    }
    return omero


@dataclass
class ZarrLevel:
    shape: DimTuple
    chunk_size: DimTuple
    dtype: np.dtype
    zarray: zarr.core.Array


def compute_level_shapes(
    lvl0shape: DimTuple, scaling: Tuple[float, float, float, float, float], nlevels: int
) -> List[DimTuple]:
    """
    Calculate all multiresolution level shapes by repeatedly scaling.
    Minimum dimension size will always be 1.
    This will always return nlevels even if the levels become unreducible and
    have to repeat.

    Parameters
    ----------
    lvl0shape:
        Shape of the array. Assumes a 5d TCZYX tuple.
    scaling:
        Amount to scale each dimension by. Dims will be DIVIDED by
        these values.
    nlevels:
        Number of levels to return. The first level is the
        original lvl0shape.

    Returns
    -------
    List of shapes of all nlevels.
    """
    shapes = [lvl0shape]
    for i in range(nlevels - 1):
        nextshape = (
            max(int(shapes[i][0] / scaling[0]), 1),
            max(int(shapes[i][1] / scaling[1]), 1),
            max(int(shapes[i][2] / scaling[2]), 1),
            max(int(shapes[i][3] / scaling[3]), 1),
            max(int(shapes[i][4] / scaling[4]), 1),
        )
        shapes.append(nextshape)
    return shapes


def get_scale_ratio(
    level0: Tuple[int, ...], level1: Tuple[int, ...]
) -> Tuple[float, ...]:
    return (
        level0[0] / level1[0],
        level0[1] / level1[1],
        level0[2] / level1[2],
        level0[3] / level1[3],
        level0[4] / level1[4],
    )


def compute_level_chunk_sizes_zslice(shapes: List[DimTuple]) -> List[DimTuple]:
    """
    Convenience function to calculate chunk sizes for each of the input level
    shapes assuming that the shapes are TCZYX and we want the chunking to be
    per Z slice.  This code also assumes we are only downsampling in XY and
    leaving TCZ alone.  For many of our microscopy images so far, we have much
    more resolution in XY than in Z so this is a reasonable assumption.

    The first shape returned will always be (1,1,1,shapes[0][3],shapes[0][4])
    and the following will be a scaled number of slices scaled by the same
    factor as the successive shapes.

    This is an attempt to keep the total size of chunks the same across all
    levels, by increasing the number of slices for downsampled levels.
    This is making a basic assumption that each of the shapes is a downsampled
    version of the previous shape.

    For example, in a typical case, if the second level is scaled down by 1/2
    in X and Y, then the second chunk size will have 4x the number of slices.

    Parameters
    ----------
    shapes:
        List of all multiresolution level shapes

    Returns
    -------
    List of chunk sizes for per-slice chunking
    """
    # assumes TCZYX order
    shape0 = shapes[0]
    chunk_sizes = []
    # assume starting with single slice
    chunk_sizes.append((1, 1, 1, shape0[3], shape0[4]))
    for i in range(1, len(shapes)):
        last_chunk_size = chunk_sizes[i - 1]
        scale = get_scale_ratio(shapes[i - 1], shapes[i])
        # assumes that scale ratio for TCZ is 1 and only downsampled in xy.
        chunk_sizes.append(
            (
                1,
                1,
                (int(scale[4] * scale[3] * last_chunk_size[2])),
                int(last_chunk_size[3] / scale[3]),
                int(last_chunk_size[4] / scale[4]),
            )
        )
    return chunk_sizes


class OmeZarrWriter:
    """
    Class to write OME-Zarr files.
    Example usage:

    .. code-block:: python

            from ome_zarr_writer import
                OmeZarrWriter,
                compute_level_shapes,
                compute_level_chunk_sizes_zslice

            # We need to compute the shapes and chunk sizes for each
            # desired multiresolution level.
            shapes = compute_level_shapes(input_shape, scaling, num_levels)
            chunk_sizes = compute_level_chunk_sizes_zslice(shapes)

            # Create an OmeZarrWriter object
            writer = OmeZarrWriter()

            # Initialize the store. Use s3 url or local directory path!
            writer.init_store(str(save_uri), shapes, chunk_sizes, im.dtype)

            # Write the image.
            # This will compute downsampled levels on the fly.
            # Adjust t batch size based on dask compute capacity.
            writer.write_t_batches_array(im, tbatch=4)

            # Generate a metadata dict and write it to the zarr.
            meta = writer.generate_metadata(
                image_name="my_image_name",
                channel_names=my_channel_names,
                physical_dims=physical_scale,
                physical_units=physical_units,
                channel_colors=my_channel_colors,
            )
            writer.write_metadata(meta)
    """

    def __init__(self) -> None:
        self.output_path: str = ""
        self.levels: List[ZarrLevel] = []
        self.store: zarr.Store = None
        self.root: zarr.hierarchy.Group = None

    def init_store(
        self,
        output_path: str,
        shapes: List[DimTuple],
        chunk_sizes: List[DimTuple],
        dtype: np.dtype,
        compressor: numcodecs.abc.Codec = default_compressor,
    ) -> None:
        """
        Initialize the store.

        Parameters
        ----------
        output_path:
            The output path. If it begins with "s3://" or "gs://", it is assumed
            to be a remote store. Credentials required to be provided externally.
        shapes:
            The shapes of the levels.
        chunk_sizes:
            The chunk sizes of the levels.
        dtype:
            The data type.
        """
        if len(shapes) != len(chunk_sizes) or len(shapes) < 1 or len(chunk_sizes) < 1:
            raise ValueError(
                "shapes and chunk_sizes must have the same length.  This is "
                "the number of multiresolution levels."
            )

        self.output_path = output_path
        # assumes authentication/permission for writes
        is_remote = output_path.startswith("s3://") or output_path.startswith("gs://")
        if is_remote:
            self.store = FSStore(url=output_path, dimension_separator="/")
        else:
            self.store = DirectoryStore(output_path, dimension_separator="/")
        # create a group with all the levels
        self.root = zarr.group(store=self.store, overwrite=True)
        # pre-create all levels here?
        self._create_levels(
            root=self.root,
            level_shapes=shapes,
            level_chunk_sizes=chunk_sizes,
            dtype=dtype,
            compressor=compressor,
        )

    def _create_levels(
        self,
        root: zarr.Group,
        level_shapes: List[DimTuple],
        level_chunk_sizes: List[DimTuple],
        dtype: np.dtype,
        compressor: numcodecs.abc.Codec = default_compressor,
    ) -> None:
        self.levels = []
        for i in range(len(level_shapes)):
            lvl = (
                root.zeros(
                    str(i),
                    shape=level_shapes[i],
                    chunks=level_chunk_sizes[i],
                    dtype=dtype,
                    compressor=compressor,
                )
                if root is not None
                else None
            )
            level = ZarrLevel(level_shapes[i], level_chunk_sizes[i], dtype, lvl)
            self.levels.append(level)

    def _downsample_and_write_batch_t(
        self, data_tczyx: da.Array, start_t: int, end_t: int, toffset: int = 0
    ) -> None:
        dtype = data_tczyx.dtype
        if len(data_tczyx.shape) != 5:
            raise ValueError("data_tczyx must be 5D")
        if len(data_tczyx) != end_t - start_t:
            raise ValueError("data_tczyx must have the same T length as end_t-start_t")

        # write level 0 first
        for k in range(start_t, end_t):
            subset = data_tczyx[[k - start_t]]
            da.to_zarr(
                subset,
                self.levels[0].zarray,
                region=(slice(k + toffset, k + toffset + 1),),
            )

        # downsample to next level then write
        for j in range(1, len(self.levels)):
            # downsample to next level
            nextshape = (end_t - start_t,) + self.levels[j].shape[1:]
            data_tczyx = resize(data_tczyx, nextshape, order=0)
            data_tczyx = data_tczyx.astype(dtype)

            # write ti to zarr
            for k in range(start_t, end_t):
                subset = data_tczyx[[k - start_t]]
                da.to_zarr(
                    subset,
                    self.levels[j].zarray,
                    region=(slice(k + toffset, k + toffset + 1),),
                )

        log.info(f"Completed {start_t} to {end_t}")

    def write_t_batches(
        self,
        im: BioImage,
        channels: List[int] = [],
        tbatch: int = 4,
        debug: bool = False,
    ) -> None:
        """
        Write the image in batches of T.

        Parameters
        ----------
        im:
            The BioImage object.
        tbatch:
            The number of T to write at a time.
        """
        # loop over T in batches
        numT = im.dims.T
        if debug:
            numT = np.min([5, numT])
        log.info("Starting loop over T")
        for i in np.arange(0, numT + 1, tbatch):
            start_t = i
            end_t = min(i + tbatch, numT)
            if end_t > start_t:
                # assume start t and end t are in range (caller should guarantee this)
                ti = im.get_image_dask_data(
                    "TCZYX", T=slice(start_t, end_t), C=channels
                )
                self._downsample_and_write_batch_t(ti, start_t, end_t)
        log.info("Finished loop over T")

    def write_t_batches_image_sequence(
        self,
        paths: List[str],
        channels: List[int] = [],
        tbatch: int = 4,
        debug: bool = False,
    ) -> None:
        """
        Write the image in batches of T.

        Parameters
        ----------
        paths:
            The list of file paths, one path per T.
        tbatch:
            The number of T to write at a time.
        """
        # loop over T in batches
        numT = len(paths)
        if debug:
            numT = np.min([5, numT])
        log.info("Starting loop over T")
        for i in np.arange(0, numT + 1, tbatch):
            start_t = i
            end_t = min(i + tbatch, numT)
            if end_t > start_t:
                # read batch into dask array
                ti = []
                for j in range(start_t, end_t):
                    im = BioImage(paths[j])
                    ti.append(im.get_image_dask_data("CZYX", C=channels))
                ti = da.stack(ti, axis=0)
                self._downsample_and_write_batch_t(ti, start_t, end_t)
        log.info("Finished loop over T")

    def write_t_batches_array(
        self,
        im: Union[da.Array, np.ndarray],
        channels: List[int] = [],
        tbatch: int = 4,
        toffset: int = 0,
        debug: bool = False,
    ) -> None:
        """
        Write the image in batches of T.

        Parameters
        ----------
        im:
            An ArrayLike object. Should be 5D TCZYX.
        tbatch:
            The number of T to write at a time.
        toffset:
            The offset to start writing T from. All T in the input array will be written
        """
        # if isinstance(im, (np.ndarray)):
        #     im_da = da.from_array(im)
        # else:
        #     im_da = im
        im_da = im
        # loop over T in batches
        numT = im_da.shape[0]
        if debug:
            numT = np.min([5, numT])
        log.info("Starting loop over T")
        for i in np.arange(0, numT + 1, tbatch):
            start_t = i
            end_t = min(i + tbatch, numT)
            if end_t > start_t:
                # assume start t and end t are in range (caller should guarantee this)
                ti = im_da[start_t:end_t]
                if channels:
                    for t in range(len(ti)):
                        ti[t] = [ti[t][c] for c in channels]
                self._downsample_and_write_batch_t(
                    da.asarray(ti), start_t, end_t, toffset
                )
        log.info("Finished loop over T")

    def _get_scale_ratio(self, level: int) -> Tuple[float, float, float, float, float]:
        lvl_shape = self.levels[level].shape
        lvl0_shape = self.levels[0].shape
        return (
            lvl0_shape[0] / lvl_shape[0],
            lvl0_shape[1] / lvl_shape[1],
            lvl0_shape[2] / lvl_shape[2],
            lvl0_shape[3] / lvl_shape[3],
            lvl0_shape[4] / lvl_shape[4],
        )

    def generate_metadata(
        self,
        image_name: str,
        channel_names: List[str],
        physical_dims: dict,  # {"x":0.1, "y", 0.1, "z", 0.3, "t": 5.0}
        physical_units: dict,  # {"x":"micrometer", "y":"micrometer",
        # "z":"micrometer", "t":"minute"},
        channel_colors: Union[List[str], List[int]],
    ) -> dict:
        """
        Build a metadata dict suitable for writing to ome-zarr attrs.

        Parameters
        ----------
        image_name:
            The image name.
        channel_names:
            The channel names.
        physical_dims:
            for each physical dimension, include a scale
            factor.  E.g. {"x":0.1, "y", 0.1, "z", 0.3, "t": 5.0}
        physical_units:
            For each physical dimension, include a unit
            string. E.g. {"x":"micrometer", "y":"micrometer", "z":"micrometer",
            "t":"minute"}
        """
        dims = ("t", "c", "z", "y", "x")
        axes = []
        for dim in dims:
            unit = None
            if physical_units and dim in physical_units:
                unit = physical_units[dim]
            if dim in {"x", "y", "z"}:
                axis = Axis(name=dim, type="space", unit=unit)
            elif dim == "c":
                axis = Axis(name=dim, type="channel", unit=unit)
            elif dim == "t":
                axis = Axis(name=dim, type="time", unit=unit)
            else:
                msg = f"Dimension identifier is not valid: {dim}"
                raise KeyError(msg)
            axes.append(axis)

        datasets = []
        for index, level in enumerate(self.levels):
            path = f"{index}"
            scale = []
            level_scale = self._get_scale_ratio(index)
            level_scale_dict = dim_tuple_to_dict(level_scale)
            for dim in dims:
                phys = (
                    physical_dims[dim] * level_scale_dict[dim]
                    if dim in physical_dims and dim in level_scale_dict
                    else 1.0
                )
                scale.append(phys)
            translation = []
            for dim in dims:
                # TODO handle optional translations e.g. xy stage position,
                # start time etc
                translation.append(0.0)
            coordinateTransformations = [Scale(scale), Translation(translation)]
            dataset = Dataset(
                path=path, coordinateTransformations=coordinateTransformations
            )
            datasets.append(dataset)
        metadata = Metadata(
            axes=axes,
            datasets=datasets,
            name="/",
            coordinateTransformations=None,
        )

        metadata_dict = asdict(metadata)
        metadata_dict = _pop_metadata_optionals(metadata_dict)

        # get the total shape as dict:
        shapedict = dim_tuple_to_dict(self.levels[0].shape)

        # add the omero data
        ome_json = build_ome(
            shapedict["z"] if "z" in shapedict else 1,
            image_name,
            channel_names=channel_names,  # assumes we have written all channels!
            channel_colors=channel_colors,  # type: ignore
            # TODO: Rely on user to supply the per-channel min/max.
            channel_minmax=[
                (0.0, 1.0) for i in range(shapedict["c"] if "c" in shapedict else 1)
            ],
        )

        ome_zarr_metadata = {"multiscales": [metadata_dict], "omero": ome_json}
        return ome_zarr_metadata

    def write_metadata(self, metadata: dict) -> None:
        """
        Write the metadata.

        Parameters
        ----------
        metadata:
            The metadata dict. Expected to contain a multiscales
            array and omero dict
        """
        self.root.attrs["multiscales"] = metadata["multiscales"]
        self.root.attrs["omero"] = metadata["omero"]
